﻿/*****************************************************************************
*                                                                            *
*  XELF.OpenNI.Sample                                                        *
*  Copyright © 2010-2011 XELF                                                *
*  http://xelf.info/                                                         *
*                                                                            *
*  This file is part of XELF.OpenNI.Sample                                   *
*                                                                            *
*  XELF.OpenNI.Sample stays under a dual license: Ms-PL or LGPL              *
*                                                                            *
*  Please note that differences between XELF.OpenNI and XELF.OpenNI.Sample   *
*  in these licences.                                                        *
*                                                                            *
*****************************************************************************/

using System.Collections.Generic;
using Microsoft.Xna.Framework;
using XELF.OpenNI;

namespace XELF.MotionDesign {

	struct Pose {
		public static readonly Pose Identity = new Pose { Rotation = Quaternion.Identity, };

		public Quaternion Rotation;
		public Microsoft.Xna.Framework.Vector3 Position;
	}

	class JointNode {
		public bool Enabled = true;

		public SkeletonJoint Current;
		public SkeletonJoint Parent;
		public SkeletonJointTransform? Transform;

		public Quaternion FilteredLocalRotation = Quaternion.Identity;

		public Pose RawWorldPose = Pose.Identity;
		public Pose WorldPose = Pose.Identity;
		public Pose LocalPose = Pose.Identity;

		readonly PoseFilter filter = new PoseFilter();

		public void UpdateTransform(UserGenerator userGenerator, ushort user) {
			Transform = userGenerator.GetSkeletonJoint(user, Current);
		}

		public void UpdateRawWorldPosition(ref Matrix world) {
			if (Transform.HasValue) {
				RawWorldPose.Position = Transform.Value.Position.Position.ToVector3();
				Microsoft.Xna.Framework.Vector3.Transform(ref RawWorldPose.Position, ref world, out RawWorldPose.Position);
			}
		}

		public void UpdateRawWorldRotation(ref Quaternion worldRotation) {
			if (Transform.HasValue) {
				var orientation = Transform.Value.Orientation.Orientation;
				if (!orientation.IsZero()) {
					XELF.OpenNI.Helper.CreateQuaternionFromOrientation(ref orientation, out RawWorldPose.Rotation);
					Quaternion.Multiply(ref worldRotation, ref RawWorldPose.Rotation, out RawWorldPose.Rotation);
				} else {
					RawWorldPose.Rotation = Quaternion.Identity;
				}
			}
		}
		public void UpdateWorldPose() {
			filter.Input(RawWorldPose);
			WorldPose = filter.Output;
		}
		public void UpdateLocalPosition(JointNodes nodes) {
			if (Parent != SkeletonJoint.None) {
				Matrix matrix;
				Matrix.CreateFromQuaternion(ref nodes[Parent].WorldPose.Rotation, out matrix);
				matrix.Translation = nodes[Parent].WorldPose.Position;
				Matrix.Invert(ref matrix, out matrix);
				Microsoft.Xna.Framework.Vector3.Transform(ref WorldPose.Position, ref matrix, out LocalPose.Position);
			} else {
				LocalPose.Position = WorldPose.Position;
			}
		}
		public void UpdateLocalRotation(JointNodes nodes) {
			if (Parent != SkeletonJoint.None) {
				Quaternion parentInverseWorldRotation;
				Quaternion.Conjugate(ref nodes[Parent].WorldPose.Rotation, out parentInverseWorldRotation);
				//Quaternion.Multiply(ref parentInverseWorldRotation, ref WorldRotation, out LocalRotation);
				Quaternion.Multiply(ref WorldPose.Rotation, ref parentInverseWorldRotation, out LocalPose.Rotation);
			} else {
				LocalPose.Rotation = WorldPose.Rotation;
			}
		}
	}

	class JointNodes {
		public void Update(UserGenerator userGenerator, ushort user, PositionNormal? plane) {
			UpdateTransform(userGenerator, user);
			UpdateLocal(userGenerator, user, plane);
		}
		void UpdateTransform(UserGenerator userGenerator, ushort user) {
			if (!userGenerator.IsSkeletonTracking(user)) return;
			foreach (var node in Nodes) {
				if (node.Enabled) {
					node.UpdateTransform(userGenerator, user);
				}
			}
		}
		void UpdateLocal(UserGenerator userGenerator, ushort user, PositionNormal? plane) {
			if (!userGenerator.IsSkeletonTracking(user)) return;

			Matrix world = Matrix.Identity;
			if (plane.HasValue) {
				var position = Microsoft.Xna.Framework.Vector3.Zero;
				var up = plane.Value.Normal.ToVector3();
				var forward = Microsoft.Xna.Framework.Vector3.Forward;
				Microsoft.Xna.Framework.Vector3 side;
				Microsoft.Xna.Framework.Vector3.Cross(ref up, ref forward, out side);
				side.Normalize();
				Microsoft.Xna.Framework.Vector3.Cross(ref side, ref up, out forward);
				forward.Normalize();
				Matrix.CreateWorld(ref position, ref forward, ref up, out world);
			}

			Quaternion worldRotation;
			Quaternion.CreateFromRotationMatrix(ref world, out worldRotation);

			foreach (var node in Nodes) {
				node.UpdateRawWorldRotation(ref worldRotation);
				node.UpdateRawWorldPosition(ref world);
			}
			foreach (var node in Nodes) {
				node.UpdateWorldPose();
			}
			foreach (var node in Nodes) {
				node.UpdateLocalRotation(this);
				node.UpdateLocalPosition(this);
			}
		}

		public JointNode this[SkeletonJoint joint] {
			get { return Nodes[(int)joint]; }
		}

		public List<JointNode> Nodes = new List<JointNode> {
			new JointNode { Enabled = false, Parent = SkeletonJoint.None, }, // None = 0,
			new JointNode { Current = SkeletonJoint.Head, Parent = SkeletonJoint.Neck, }, // Head = 1,
			new JointNode { Current = SkeletonJoint.Neck, Parent = SkeletonJoint.Torso, }, // Neck = 2,
			new JointNode { Current = SkeletonJoint.Torso, Parent = SkeletonJoint.None, }, // Torso = 3,
			new JointNode { Enabled =false, Current = SkeletonJoint.Waist, }, // Waist = 4,
			new JointNode { Enabled =false, Current = SkeletonJoint.LeftCollar, }, // LeftCollar = 5,
			new JointNode { Current = SkeletonJoint.LeftShoulder, Parent = SkeletonJoint.Torso, }, // LeftShoulder = 6,
			new JointNode { Current = SkeletonJoint.LeftElbow, Parent = SkeletonJoint.LeftShoulder, }, // LeftElbow = 7,
			new JointNode { Enabled =false,Current = SkeletonJoint.LeftWrist, }, // LeftWrist = 8,
			new JointNode { Current = SkeletonJoint.LeftHand, Parent = SkeletonJoint.LeftElbow, }, // LeftHand = 9,
			new JointNode { Enabled =false, Current = SkeletonJoint.LeftFingertip, }, // LeftFingertip = 10,
			new JointNode { Enabled =false, Current = SkeletonJoint.RightCollar, }, // RightCollar = 11,
			new JointNode { Current = SkeletonJoint.RightShoulder, Parent = SkeletonJoint.Torso, }, // RightShoulder = 12,
			new JointNode { Current = SkeletonJoint.RightElbow, Parent = SkeletonJoint.RightShoulder, }, // RightElbow = 13,
			new JointNode { Enabled =false, Current = SkeletonJoint.RightWrist, }, // RightWrist = 14,
			new JointNode { Current = SkeletonJoint.RightHand, Parent = SkeletonJoint.RightElbow, }, // RightHand = 15,
			new JointNode { Enabled =false, Current = SkeletonJoint.RightFingertip, }, // RightFingertip = 16,
			new JointNode { Current = SkeletonJoint.LeftHip, Parent = SkeletonJoint.None, }, // LeftHip = 17,
			new JointNode { Current = SkeletonJoint.LeftKnee, Parent = SkeletonJoint.LeftHip, }, // LeftKnee = 18,
			new JointNode { Enabled =false, Current = SkeletonJoint.LeftAnkle, }, // LeftAnkle = 19,
			new JointNode { Current = SkeletonJoint.LeftFoot, Parent = SkeletonJoint.LeftKnee, }, // LeftFoot = 20,
			new JointNode { Current = SkeletonJoint.RightHip, Parent = SkeletonJoint.None, }, // RightHip = 21,
			new JointNode { Current = SkeletonJoint.RightKnee, Parent = SkeletonJoint.RightHip, }, // RightKnee = 22,
			new JointNode { Enabled =false, Current = SkeletonJoint.RightAnkle, }, // RightAnkle = 23,
			new JointNode { Current = SkeletonJoint.RightFoot, Parent = SkeletonJoint.RightKnee, }, // RightFoot = 24,
        };
	}

}
