﻿using System;
using System.Threading;

namespace XELF.OpenNI {

	public class OpenNIDevice : IDisposable {
		public Context Context { get; private set; }
		public readonly ErrorCollection Errors = new ErrorCollection();
		public DepthGenerator DepthGenerator { get; private set; }
		public ImageGenerator ImageGenerator { get; private set; }
		public UserGenerator UserGenerator { get; private set; }
		public SceneAnalyzer SceneAnalyzer { get; private set; }

		public MapOutputMode Mode = new MapOutputMode {
			Width = 640,
			Height = 480,
			FramesPerSecond = 30,
		};

		#region Statistics fields

		public int UserCalibrationSucceededCount;
		public int UserCalibrationFailedCount;

		#endregion

		public void Dispose() {
			exiting = true;
			if (thread != null) {
				thread.Join();
				thread = null;
			}
			if (Context != null) {
				Context.Dispose();
				Context = null;
			}
		}

		public void Initialize() {
			Dispose();
			exiting = false;

			// FIXME:
			for (int i = 0; i < 10; i++) {
				try {
					Context = new Context("config.xml", null);
				} catch {
					foreach (var error in Errors) {
						//
					}
					continue;
				}
				break;
			}

			if (Context == null) {
				foreach (var error in Errors) {
					Console.WriteLine(error.ToString());
				}
				throw new Exception("OpenNI Initalization Failed.");
			}
			//depthGenerator = DepthGenerator.Find(context);
			DepthGenerator = new DepthGenerator(Context);
			//imageGenerator = ImageGenerator.Find(context);
			ImageGenerator = new ImageGenerator(Context);
			UserGenerator = UserGenerator.Find(Context);
			//userGenerator = new UserGenerator(context, errors);

			var canFitViewPoint = DepthGenerator.IsCapabilitySupported(Capability.AlternativeViewPoint)
				&& DepthGenerator.IsViewPointSupported(ImageGenerator);
			if (canFitViewPoint) {
				DepthGenerator.ViewPoint = ImageGenerator;
			}

			UserGenerator.SetSkeletonSmoothing(0);
			UserGenerator.UserEntered += new Action<ushort>(UserEntered);
			UserGenerator.UserLeft += new Action<ushort>(UserLeft);
			UserGenerator.CalibrationStart += new Action<ushort>(UserCalibrationStart);
			UserGenerator.CalibrationEnd += new Action<ushort, bool>(UserCalibrationEnd);

			SceneAnalyzer = new SceneAnalyzer(Context, Errors);

			ImageGenerator.MapOutputMode = Mode;
			DepthGenerator.MapOutputMode = Mode;

			Context.Check(Context.Start());

			thread = new Thread(new ThreadStart(ThreadMain));
			thread.Start();
		}

		#region Thread

		Thread thread;

		volatile bool exiting;

		void ThreadMain() {
			while (!exiting) {
				Context.WaitAndUpdateAll();
			}
		}

		#endregion

		void UserEntered(ushort user) {
			UserGenerator.RequestSkeletonCalibration(user, false);
		}
		void UserLeft(ushort user) {
		}
		void UserCalibrationStart(ushort user) {
		}
		void UserCalibrationEnd(ushort user, bool success) {
			if (success) {
				UserCalibrationSucceededCount++;
				UserGenerator.StartSkeletonTracking(user);

				for (SkeletonJoint joint = SkeletonJoint.Head; joint <= SkeletonJoint.RightFoot; joint++) {
					try {
						UserGenerator.SetJointActive(joint, true);
					} catch {
					}
				}
			} else {
				UserGenerator.RequestSkeletonCalibration(user, true);
				UserCalibrationFailedCount++;
			}
		}

	}

}
