﻿/*****************************************************************************
*                                                                            *
*  XELF.OpenNI.Sample                                                        *
*  Copyright © 2010-2011 XELF                                                *
*  http://xelf.info/                                                         *
*                                                                            *
*  This file is part of XELF.OpenNI.Sample                                   *
*                                                                            *
*  XELF.OpenNI.Sample stays under a dual license: Ms-PL or LGPL              *
*                                                                            *
*  Please note that differences between XELF.OpenNI and XELF.OpenNI.Sample   *
*  in these licences.                                                        *
*                                                                            *
*****************************************************************************/

using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using XELF.OpenNI;

namespace XELF.MotionDesign {

	public class OpenNIGame : Microsoft.Xna.Framework.Game {
		#region Graphics fields
		GraphicsDeviceManager graphics;
		SpriteBatch spriteBatch;
		Texture2D pointTexture;
		SpriteFont fontSmall;
		SpriteFont fontBig;
		readonly StringBuilder text = new StringBuilder();
		#endregion

		#region Settings fields
		bool StatusVisible = true;
		bool PointVisible = true;
		bool ImageVisible = true;
		bool DepthVisible = true;
		#endregion

		KeyboardState previousKeyboardState;
		KeyboardState currentKeyboardState;

		#region Low level fields

		readonly OpenNIDevice ni = new OpenNIDevice();

		PositionNormal? floorPlane;
		ushort[] users = new ushort[16];

		OpenNI.Vector3[] projectivePoints = new OpenNI.Vector3[25];
		OpenNI.Vector3[] realWorldPoints = new OpenNI.Vector3[25];

		int numberActiveJoints;
		SkeletonJoint[] activeJoints = new SkeletonJoint[24];

		ImageColorTextureGenerator imageTextureGenerator;
		DepthColorTextureGenerator depthTextureGenerator;

		#endregion

		#region Middle level fields

		readonly JointNodes jointNodes = new JointNodes();

		#endregion

		public OpenNIGame() {
			graphics = new GraphicsDeviceManager(this) {
				PreferredBackBufferWidth = 640,
				PreferredBackBufferHeight = 480,
				PreferMultiSampling = true,
			};
			Content.RootDirectory = "Content";
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			fontSmall = Content.Load<SpriteFont>("SpriteFontSmall");
			fontBig = Content.Load<SpriteFont>("SpriteFontBig");
			pointTexture = new Texture2D(GraphicsDevice, 1, 1);
			pointTexture.SetData(new[] { Color.White });

			ni.Initialize();
			imageTextureGenerator = new ImageColorTextureGenerator(
				ni.ImageGenerator, new Texture2D(GraphicsDevice, (int)ni.Mode.Width, (int)ni.Mode.Height));
			depthTextureGenerator = new DepthColorTextureGenerator(
				ni.DepthGenerator, new Texture2D(GraphicsDevice, (int)ni.Mode.Width, (int)ni.Mode.Height));
		}

		void UpdateJoints() {
			if (!ni.UserGenerator.IsSkeletonTracking(users[0])) return;
			for (int i = 0; i < numberActiveJoints; i++) {
				var joint = ni.UserGenerator.GetSkeletonJoint(users[0], activeJoints[i]);
				if (joint != null) {
					realWorldPoints[(int)activeJoints[i]] = joint.Value.Position.Position;
				}
			}
			ni.DepthGenerator.ConvertRealWorldToProjective(realWorldPoints, realWorldPoints.Length, projectivePoints);
		}

		bool IsKeyPressed(Keys key) {
			return currentKeyboardState.IsKeyDown(key) && previousKeyboardState.IsKeyUp(key);
		}

		protected override void Update(GameTime gameTime) {
			previousKeyboardState = currentKeyboardState;
			currentKeyboardState = Keyboard.GetState();

			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();

			if (IsKeyPressed(Keys.Enter)) {
				var p = GraphicsDevice.PresentationParameters;
				p.IsFullScreen = !p.IsFullScreen;
				GraphicsDevice.Reset(p);
			}
			if (IsKeyPressed(Keys.I)) {
				ImageVisible = !ImageVisible;
			}
			if (IsKeyPressed(Keys.D)) {
				DepthVisible = !DepthVisible;
			}
			if (IsKeyPressed(Keys.S)) {
				StatusVisible = !StatusVisible;
			}
			if (IsKeyPressed(Keys.P)) {
				PointVisible = !PointVisible;
			}

			if (IsKeyPressed(Keys.Space)) {
				try {
					ni.UserGenerator.StartSkeletonTracking(users[0]);
				} catch {
					ni.UserGenerator.RequestSkeletonCalibration(users[0], true);
				}
			}

			ni.UserGenerator.SetSkeletonSmoothing(0.5f);

			ni.UserGenerator.GetUsers(users);
			floorPlane = ni.SceneAnalyzer.Floor;

			numberActiveJoints = ni.UserGenerator.GetActiveJoints(activeJoints);

			UpdateJoints();

			jointNodes.Update(ni.UserGenerator, users[0], floorPlane);

			base.Update(gameTime);
		}

		protected override void Draw(GameTime gameTime) {
			imageTextureGenerator.Update();
			depthTextureGenerator.Update();

			GraphicsDevice.Clear(Color.White);

			spriteBatch.Begin();

			if (DepthVisible)
				spriteBatch.Draw(depthTextureGenerator.Texture, Vector2.Zero, Color.White);
			if (ImageVisible)
				spriteBatch.Draw(imageTextureGenerator.Texture, Vector2.Zero, new Color(1, 1, 1, 0.5f));

			if (PointVisible) {
				for (int i = 0; i < projectivePoints.Length; i++) {
					var x = projectivePoints[i].X;
					var y = projectivePoints[i].Y;
					spriteBatch.Draw(pointTexture, new Vector2(x, y), null, Color.Red, 0, new Vector2(0.5f, 0.5f), 32.0f, SpriteEffects.None, 0);
				}
			}

			if (StatusVisible) {
				text.Clear();

				text.AppendFormat(
					"number users = {0}\nusers[0] {{ id = {1}\n / center of mass = {2} }}\nplane = {{{3}}}\n{4}\nDepth Output: {5}\n",
					ni.UserGenerator.Count, users[0], ni.UserGenerator.GetUserCenterOfMass(users[0]),
					floorPlane, realWorldPoints[2], ni.DepthGenerator.MapOutputMode);

				text.Append("Active Joints:\n");
				if (ni.UserGenerator.IsSkeletonTracking(users[0])) {
					for (int i = 0; i < numberActiveJoints; i++) {
						text.Append(activeJoints[i].ToString());
						var joint = ni.UserGenerator.GetSkeletonJoint(users[0], activeJoints[i]);
						if (joint != null) {
							var wp = joint.Value.Position.Position;
							var orientation = joint.Value.Orientation.Orientation;
							text.AppendFormat(": P=({0:+0000;-0000} {1:+0000;-0000} {2:+0000;-0000}) ({3:+0;-0})",
								wp.X, wp.Y, wp.Z, joint.Value.Position.Confidence);
							//text.AppendFormat(", R=[{0:#0.##} {1:#0.##} {2:#0.##}][{3:#0.##} {4:#0.##} {5:#0.##}][{6:#0.##} {7:#0.##} {8:#0.##}] ({9})",
							//     orientation.M11, orientation.M12, orientation.M13,
							//     orientation.M21, orientation.M22, orientation.M23,
							//     orientation.M31, orientation.M32, orientation.M33, joint.Value.Orientation.Confidence);
							Quaternion wq;
							XELF.OpenNI.Helper.CreateQuaternionFromOrientation(ref orientation, out wq);
							text.AppendFormat(", WQ=[({0:+0.00;-0.00}, {1:+0.00;-0.00}, {2:+0.00;-0.00}), {3:+0.00;-0.00}] |WQ|={4} ({5})", wq.X, wq.Y, wq.Z, wq.W, wq.Length(), joint.Value.Confidence);
							var lq = jointNodes[activeJoints[i]].LocalPose.Rotation;
							text.AppendFormat(", LQ=[({0:+0.00;-0.00}, {1:+0.00;-0.00}, {2:+0.00;-0.00}), {3:+0.00;-0.00}]", lq.X, lq.Y, lq.Z, lq.W);
						}
						text.Append(",\n");
					}
				}

				spriteBatch.DrawString(fontSmall, text, new Vector2(10, 10), Color.Blue);

				text.Clear();
				text.AppendFormat("users[0] Tracking: {0}\n",
					ni.UserGenerator.IsSkeletonTracking(users[0]));
				text.AppendFormat("Calibration: Succeeded={0} / Failed={1}\n",
					ni.UserCalibrationSucceededCount, ni.UserCalibrationFailedCount);

				spriteBatch.DrawString(fontBig, text, new Vector2(10, 400), Color.Blue);

			}

			spriteBatch.End();

			base.Draw(gameTime);
		}

		protected override void Dispose(bool disposing) {
			ni.Dispose();
			base.Dispose(disposing);
		}
	}
}
