﻿/*****************************************************************************
*                                                                            *
*  XELF.OpenNI.Sample                                                        *
*  Copyright © 2010-2011 XELF                                                *
*  http://xelf.info/                                                         *
*                                                                            *
*  This file is part of XELF.OpenNI.Sample                                   *
*                                                                            *
*  XELF.OpenNI.Sample stays under a dual license: Ms-PL or LGPL              *
*                                                                            *
*  Please note that differences between XELF.OpenNI and XELF.OpenNI.Sample   *
*  in these licences.                                                        *
*                                                                            *
*****************************************************************************/

using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XELF.OpenNI {

	public abstract class TextureGenerator<RawT, ImageT>
		where ImageT : struct {
		
		protected Generator Generator;
		RawT[] buffer;
		ImageT[] image;
		public Texture2D Texture { get; private set; }

		public TextureGenerator(Generator generator, Texture2D target, int rawElementsPerImageElement) {
			Generator = generator;
			Texture = target;
			var length = target.Width * target.Height;
			buffer = new RawT[length * rawElementsPerImageElement];
			image = new ImageT[length];
		}

		protected abstract void UpdateMap(RawT[] destinationBuffer);
		protected abstract void ConvertImage(RawT[] sourceBuffer, ImageT[] destinationImage);

		public void Update() {
			UpdateMap(buffer);
			ConvertImage(buffer, image);
			Texture.SetData(image);
		}
	}

	public class ImageColorTextureGenerator : TextureGenerator<byte, Color> {
		public ImageGenerator ImageGenerator { get; private set; }

		public ImageColorTextureGenerator(ImageGenerator generator, Texture2D target)
			: base(generator, target, 3) {
			ImageGenerator = generator;
		}
		protected override void UpdateMap(byte[] destinationBuffer) {
			Marshal.Copy(ImageGenerator.Map, destinationBuffer, 0, destinationBuffer.Length);
		}
		protected override void ConvertImage(byte[] sourceBuffer, Color[] destinationImage) {
			for (int i = 0; i < destinationImage.Length; i++) {
				var r = sourceBuffer[i * 3];
				var g = sourceBuffer[i * 3 + 1];
				var b = sourceBuffer[i * 3 + 2];
				destinationImage[i] = new Color(r, g, b);
			}
		}
	}

	public class DepthColorTextureGenerator : TextureGenerator<short, Color> {
		public DepthGenerator DepthGenerator { get; private set; }

		public DepthColorTextureGenerator(DepthGenerator generator, Texture2D target)
			: base(generator, target, 1) {
			DepthGenerator = generator;
		}
		protected override void UpdateMap(short[] target) {
			Marshal.Copy(DepthGenerator.Map, target, 0, target.Length);
		}
		protected override void ConvertImage(short[] source, Color[] destination) {
			for (int i = 0; i < destination.Length; i++) {
				var z = (ushort)source[i] >> 3;
				destination[i] = new Color(z, z, z);
			}
		}
	}

	public static partial class Helper {
		public static Microsoft.Xna.Framework.Vector3 ToVector3(this OpenNI.Vector3 value) {
			return new Microsoft.Xna.Framework.Vector3(value.X, value.Y, value.Z);
		}
		public static bool IsZero(this Matrix3x3 value) {
			return
				value.M11 == 0 &&
				value.M12 == 0 &&
				value.M13 == 0 &&
				value.M21 == 0 &&
				value.M22 == 0 &&
				value.M23 == 0 &&
				value.M31 == 0 &&
				value.M32 == 0 &&
				value.M33 == 0;
		}
		public static void CreateMatrixFromOrientation(ref Matrix3x3 matrix3x3, out Matrix result) {
			result = new Matrix(
				matrix3x3.M11, matrix3x3.M12, matrix3x3.M13, 0,
				matrix3x3.M21, matrix3x3.M22, matrix3x3.M23, 0,
				matrix3x3.M31, matrix3x3.M32, matrix3x3.M33, 0,
				0, 0, 0, 1);
		}
		public static void CreateQuaternionFromOrientation(ref Matrix3x3 matrix3x3, out Quaternion result) {
			Matrix matrix;
			CreateMatrixFromOrientation(ref matrix3x3, out matrix);
			Quaternion.CreateFromRotationMatrix(ref matrix, out result);
		}
	}

}
